// A small external to interpret the Wiimote movement output
// Not really need, but could be handy
//
// Usage: should explain itself. Output depends on creation argument.
//        possible options are: raw, total, pitch or roll.
//
// ===================================================================
// WiiAcc external
// Part of the Wiilib library for Puredata
// Written by Yvan Vander Sanden (Jan 2008), http://youngmusic.org
//
// Requires the CWiid library (version 0.6.00) by L. Donnie Smith
//
// ===================================================================
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
// ===================================================================
//

#include <m_pd.h>
#include <cwiid.h>
#include <math.h>
#include <glib.h>
#include "wiiAcc.h"

#define PI 3.14159265358979323

void wiiAcc_post(t_wiiAcc *x) {
	post("highest: %f", x->highest);
	post("lowest: %f", x->lowest);
}

t_float wiiAcc_sort(gpointer a, gpointer b) {
	t_float* x = (t_float*)a;
	t_float* y = (t_float*)b;
	return *x - *y;
}

void wiiAcc_input(t_wiiAcc *x, t_symbol *s, int argc, t_atom *argv) {

	// input is a list with 3 floats, the x- y- and z-axis.
	x->x = atom_getfloat(argv);
	x->y = atom_getfloat(argv+1);
	x->z = atom_getfloat(argv+2);

	// raw output is pretty straitforward. But it saves an unpack box and 
	// makes this a bit more consistent
	//
	if (x->raw) {
		outlet_float(x->outlet_x, x->x);
		outlet_float(x->outlet_y, x->y);
		outlet_float(x->outlet_z, x->z);
	}

	// Like Mike Wozniewki wrote in the help patch that came with the original
	// external: The pitch and roll are only accurate if there are no extra
	// accelerations due to hand movement. We can check if the total acceleration
	// is close to gravity and only use pitch and roll in that case.
	if (x->total) {
		x->total_value = sqrt(pow(x->x,2) + pow(x->y,2) + pow(x->z,2));
		outlet_float(x->outlet_x, x->total_value);
	}

	if (x->pitch) {
		if (x->z != 0) {
			x->pitch_value = atan(x->y / x->z);
			if (x->z <= 0.0) x->pitch_value += PI * ((x->y > 0.0) ? 1 : -1);
			x->pitch_value *= -1;
			if (x->do_rescale) {
				x->pitch_value = x->pitch_value / PI * x->rescale;
			}
			if (x->trigger > 0) {
				if ((fabs(x->pitch_value - x->previous_pitch)) >= x->trigger) {
					outlet_float(x->outlet_x, x->pitch_value);
					x->previous_pitch = x->pitch_value;
				}
			} else {
				outlet_float(x->outlet_x, x->pitch_value);
			}
		}
	}
	if (x->roll) {
		if (x->z != 0) {
			x->roll_value = atan(x->x / x->z);
			if (x->z <= 0.0) x->roll_value += PI * ((x->x > 0.0) ? 1 : -1);
			x->roll_value *= -1;
			if (x->do_rescale) {
				x->roll_value = x->roll_value / PI * x->rescale;
			}
			if (x->trigger > 0) {
				if ((fabs(x->roll_value - x->previous_roll)) >= x->trigger) {
					outlet_float(x->outlet_x, x->roll_value);
					x->previous_roll = x->roll_value;
				}
			} else {
				outlet_float(x->outlet_x, x->roll_value);
			}
		}
	}
	if (x->horizontal) {
		x->horizontal_value = x->x;
		if (x->do_rescale) {
			x->horizontal_value =  x->x / PI * x->rescale;
		}
		if (x->trigger > 0) {
			if (x->median->len < x->trigger) {
				x->median = g_array_append_val(x->median, x->horizontal_value);
			} else {
				t_int count = x->median->len;
				g_array_sort(x->median, (GCompareFunc)wiiAcc_sort);
				if ((count % 2) != 0) {
					count /= 2;
					count += 1;
					x->horizontal_value = g_array_index(x->median, t_float, count);
				} else {
					count /= 2;
					t_float first, second;
					first = g_array_index(x->median, t_float, count);
					second = g_array_index(x->median, t_float, count+1);
					x->horizontal_value = (first + second) / 2;
					g_array_free(x->median, TRUE);
					x->median = g_array_new(TRUE, FALSE, sizeof(t_float*));
				}

				outlet_float(x->outlet_x, x->horizontal_value);
			}
		} else {
			outlet_float(x->outlet_x, x->horizontal_value);
		}
	}
	if (x->vertical) {
                x->vertical_value = x->x;
                if (x->do_rescale) {
	                x->vertical_value =  x->x / PI * x->rescale;
                }
                if (x->trigger > 0) {
                       if ((fabs(x->vertical_value - x->previous_vertical)) >= x->trigger) {
                                outlet_float(x->outlet_x, x->vertical_value);
                                x->previous_vertical = x->vertical_value;
                        }
                } else {
                        outlet_float(x->outlet_x, x->vertical_value);
                }
	}
}

void *wiiAcc_new(t_symbol *s, int argc, t_atom *argv) {
	t_wiiAcc *x = (t_wiiAcc *)pd_new(wiiAcc_class);
	x->raw = 0;
	x->roll = 0;
	x->pitch = 0;
	x->total = 0;
	x->horizontal = 0;
	x->vertical = 0;
	x->horizontal_value = 0;
	x->vertical_value = 0;
	// temp
	x->highest = 0;
	x->lowest = 0;

	// we can recalculate the output values to another scale
	x->rescale = 0;
	x->do_rescale = 0;

	// only trigger output if it differs at least this amount
	// from the previous output
	x->trigger = 0;
	x->previous_pitch = 0;
	x->previous_horizontal = 0;
	x->previous_vertical = 0;	
	// First outlet is used in all cases. 
	// Outlet 2 and 3 are only used in raw mode,
	// hence only created if needed.
	//
	x->outlet_x = outlet_new(&x->x_obj, &s_float);

	t_symbol *temp;
	if (argc == 0) {
		temp = gensym("raw");
	} else {
		temp = atom_getsymbol(argv);
	}
	if (temp == gensym("raw")) {
		x->raw = 1;
		x->outlet_y = outlet_new(&x->x_obj, &s_float);
		x->outlet_z = outlet_new(&x->x_obj, &s_float);
	}
	if (temp == gensym("roll")) {
		x->roll = 1;
	}
	if (temp == gensym("pitch")) {
		x->pitch = 1;
	}
	if (temp == gensym("total")) {
		x->total = 1;
	}

	if (temp == gensym("horizontal")) {
		x->horizontal = 1;
	}

	if (temp == gensym("vertical")) {
		x->vertical = 1;
	}

	if (argc > 1) {
		x->rescale = atom_getfloat(argv+1);
		x->do_rescale = 1;
	}
	
	if (argc > 2) {
		x->trigger = atom_getfloat(argv+2);
		// setup array for median if horizontal or vertical is selected
		if(x->vertical || x->horizontal) {
			x->median = g_array_new(TRUE, FALSE, sizeof(t_float));
		}
	}
	return (void *)x;

}

void wiiAcc_free(t_wiiAcc *x) {
	g_array_free(x->median, TRUE);
}

void wiiAcc_setup(void) {
	wiiAcc_class = class_new(gensym("wiiAcc"),
		(t_newmethod)wiiAcc_new,
		(t_method)wiiAcc_free, sizeof(t_wiiAcc),
		CLASS_DEFAULT, A_GIMME, 0);
	class_addlist(wiiAcc_class, (t_method)wiiAcc_input);
	class_addmethod(wiiAcc_class, (t_method)wiiAcc_post, gensym("post"), 0);
}
